"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const rxjs_1 = require("rxjs");
const createProxyChain_1 = require("./createProxyChain");
const buildNotificationChain_1 = require("./buildNotificationChain");
const composeParentsChain_1 = require("./composeParentsChain");
const customTraps = Symbol("_customTraps");
const observedObjects = Symbol("_observedObjects");
const weakParents = Symbol("wkp");
class _ObservableObject {
    constructor(from = {}, optHandlers = {}) {
        Object.defineProperties(this, {
            [observedObjects]: {
                value: {},
                writable: false,
                configurable: false,
                enumerable: false
            },
            [customTraps]: {
                value: {},
                writable: false,
                configurable: false,
                enumerable: false,
            },
            [weakParents]: {
                value: new WeakMap(),
                writable: false,
                configurable: false,
                enumerable: false
            }
        });
        if (optHandlers.set) {
            this[customTraps].set = optHandlers.set;
        }
        if (optHandlers.deleteProperty) {
            this[customTraps].deleteProperty = optHandlers.deleteProperty;
        }
        const handlers = Object.assign(optHandlers, {
            set: (obj, prop, value, receiver) => {
                const rawChains = Array.from(this[weakParents].get(receiver) || []);
                const objectChains = composeParentsChain_1.composeParentsChains(prop, rawChains);
                const notificationChain = buildNotificationChain_1.buildNotificationChain(obj[prop], value, ...objectChains);
                if (typeof value === "object") {
                    if (this[customTraps].set && this[customTraps].set(obj, prop, value, receiver) === false) {
                        return false;
                    }
                    /**
                     * When a property is set with an Object, we
                     * create for it a Proxy-chain (itself included).
                     * Each object will have a record in the root object
                     * weakMap, so we can easily access to the chains
                     * to reach that object (and to create notifications).
                     * Each chain is composed like "a.b.c".
                     */
                    obj[prop] = createProxyChain_1.createProxyChain(value, handlers, { all: this[weakParents] }, objectChains);
                }
                else {
                    /*
                     * We finalize the path of the keys passed in the above condition
                     * to reach “object endpoint” (like "w" for the prev. example)
                     * The path keys composition, let us subscribe to observables
                     * with dot notation like x.y.z.w
                     */
                    if (obj[prop] === value) {
                        /*
                         * If the value is the same, we return true.
                         * This cannot be considered as a fail. Also, failing would bring
                         * a strict-mode script to throw a TypeError.
                         */
                        return true;
                    }
                    if (this[customTraps].set && this[customTraps].set(obj, prop, value, receiver) === false) {
                        return false;
                    }
                    obj[prop] = value;
                }
                fireNotifications(this, notificationChain || {});
                return true;
            },
            deleteProperty: (target, prop) => {
                const rawChains = Array.from(this[weakParents].get(target) || []);
                const objectChains = composeParentsChain_1.composeParentsChains(prop, rawChains);
                const notificationChain = buildNotificationChain_1.buildNotificationChain(target[prop], undefined, ...objectChains);
                // What if customTrap returns undefined or null?
                const deleted = Boolean((this[customTraps].deleteProperty || Reflect.deleteProperty)(target, prop));
                if (deleted) {
                    fireNotifications(this, notificationChain);
                }
                return deleted;
            }
        });
        return new Proxy(Object.assign(this, createProxyChain_1.createProxyChain(from, handlers, { all: this[weakParents] })), handlers);
    }
    /**
     * Registers a custom property to be observed.
     *
     * @param {string} prop - The property or object
     * 		property to subscribe to (e.g. `epsilon`
     * 		or `time.current`)
     */
    observe(prop) {
        if (!this[observedObjects][prop]) {
            this[observedObjects][prop] = new rxjs_1.Subject();
        }
        return this[observedObjects][prop].asObservable();
    }
    /**
     * Returns the current image of a key of the main
     * object or a nested key.
     *
     * @param {string} path - dotted-notation path ("a.b.c")
     * @returns {any | undefined} - the whole observed object or part of it.
     * 	Undefined if the path is not matched;
     */
    snapshot(path) {
        if (!(path && typeof path === "string")) {
            const snapshot = Object.assign({}, this);
            return withoutProxy(snapshot);
        }
        const snapshot = path.split(".").reduce((acc, current) => {
            if (!(acc && typeof acc === "object" && !Array.isArray(acc) && current && acc.hasOwnProperty(current))) {
                // if the previous iteration returned undefined,
                // we'll forward this until the end of the loop.
                return undefined;
            }
            return acc[current];
        }, this);
        if (typeof snapshot === "object") {
            return Object.assign({}, withoutProxy(snapshot));
        }
        return snapshot;
    }
}
exports._ObservableObject = _ObservableObject;
/**
 * Calls next on RxJS Subject
 * with the current value
 * for each element in the chain
 * @param notificationChain
 * @private
 */
function fireNotifications(self, notificationChain) {
    const keys = Object.keys(notificationChain);
    for (let i = keys.length; i > 0;) {
        const key = keys[--i];
        const value = notificationChain[key];
        if (self[observedObjects][key]) {
            self[observedObjects][key].next(value);
        }
    }
}
/**
 * Iterates through a Proxy object chain and composes
 * a new object without proxy handler
 * @param object
 */
function withoutProxy(object) {
    if (typeof object !== "object") {
        return object;
    }
    const objectKeys = Object.keys(object);
    const finalObject = {};
    for (let i = objectKeys.length, prop; prop = objectKeys[--i];) {
        if (typeof object[prop] === "object") {
            Object.assign(finalObject, object, { [prop]: withoutProxy(object[prop]) });
        }
        else {
            finalObject[prop] = object[prop];
        }
    }
    return finalObject;
}
